from typing import Optional, Tuple, override

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["Inception", "inception_v1", "inception_v3", "inception_v4"]


class Inception(nn.Module):
    def __init__(self, num_classes: int, use_aux: bool | None = True) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.use_aux = use_aux

    @override
    def forward(self, x: Tensor) -> Tuple[Tensor | None, ...]:
        return super().forward(x)


class _InceptionModule(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels_1x1: int,
        reduce_3x3: int,
        out_channels_3x3: int,
        reduce_5x5: int,
        out_channels_5x5: int,
        out_channels_pool: int,
    ) -> None:
        super().__init__()

        self.branch1 = nn.ConvBNReLU2d(in_channels, out_channels_1x1, kernel_size=1)
        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, reduce_3x3, kernel_size=1),
            nn.ConvBNReLU2d(reduce_3x3, out_channels_3x3, kernel_size=3, padding=1),
        )
        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, reduce_5x5, kernel_size=1),
            nn.ConvBNReLU2d(reduce_5x5, out_channels_5x5, kernel_size=5, padding=2),
        )
        self.branch4 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.ConvBNReLU2d(in_channels, out_channels_pool, kernel_size=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class _InceptionAux(nn.Module):
    def __init__(
        self, in_channels: int, num_classes: int, pool_size: tuple[int, int]
    ) -> None:
        super().__init__()

        self.avgpool = nn.AdaptiveAvgPool2d(pool_size)
        self.conv = nn.ConvBNReLU2d(in_channels, 128, kernel_size=1)

        self.fc1 = nn.Linear(128 * pool_size[0] * pool_size[1], 1024)
        self.fc2 = nn.Linear(1024, num_classes)

        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.7)

    def forward(self, x: Tensor) -> Tensor:
        x = self.avgpool(x)
        x = self.conv(x)

        x = x.reshape(x.shape[0], -1)
        x = self.fc1(x)
        x = self.dropout(self.relu(x))
        x = self.fc2(x)

        return x


class Inception_V1(Inception):
    def __init__(self, num_classes: int = 1000, use_aux: bool = True) -> None:
        super().__init__(num_classes, use_aux)
        in_channels = 3

        self.conv1 = nn.ConvBNReLU2d(
            in_channels, 64, kernel_size=7, stride=2, padding=3
        )
        self.maxpool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.conv2 = nn.Sequential(
            nn.ConvBNReLU2d(64, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 192, kernel_size=3, stride=1, padding=1),
        )
        self.maxpool2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.incep_3 = nn.Sequential(
            _InceptionModule(192, 64, 96, 128, 16, 32, 32),
            _InceptionModule(256, 128, 128, 192, 32, 96, 64),
        )
        self.maxpool3 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.incep_4a = _InceptionModule(480, 192, 96, 208, 16, 48, 64)
        self.incep_4bcd = nn.Sequential(
            _InceptionModule(512, 160, 112, 224, 24, 64, 64),
            _InceptionModule(512, 128, 128, 256, 24, 64, 64),
            _InceptionModule(512, 112, 144, 288, 32, 64, 64),
        )
        self.incep_4e = _InceptionModule(528, 256, 160, 320, 32, 128, 128)
        self.maxpool4 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.incep_5 = nn.Sequential(
            _InceptionModule(832, 256, 160, 320, 32, 128, 128),
            _InceptionModule(832, 384, 192, 384, 48, 128, 128),
        )

        if use_aux:
            self.aux1 = _InceptionAux(512, num_classes, pool_size=(4, 4))
            self.aux2 = _InceptionAux(528, num_classes, pool_size=(4, 4))
        else:
            self.aux1 = None
            self.aux2 = None

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=0.4)
        self.fc = nn.Linear(1024, num_classes)

    def forward(self, x: Tensor) -> Tuple[Tensor, Optional[Tensor], Optional[Tensor]]:
        x = self.maxpool1(self.conv1(x))
        x = self.maxpool2(self.conv2(x))

        x = self.maxpool3(self.incep_3(x))
        x = self.incep_4a(x)
        if self.aux1 is not None and self.training:
            aux1 = self.aux1(x)
        else:
            aux1 = None

        x = self.incep_4bcd(x)
        if self.aux2 is not None and self.training:
            aux2 = self.aux2(x)
        else:
            aux2 = None

        x = self.maxpool4(self.incep_4e(x))
        x = self.avgpool(self.incep_5(x))

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x, aux2, aux1


class _InceptionModule_V2A(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 96, kernel_size=3, padding=1),
            nn.ConvBNReLU2d(96, 96, kernel_size=3, padding=1),
        )

        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 48, kernel_size=1),
            nn.ConvBNReLU2d(48, 64, kernel_size=3, padding=1),
        )

        self.branch3 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
        )

        self.branch4 = nn.ConvBNReLU2d(in_channels, 64, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class _InceptionModule_V2B(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, out_channels, kernel_size=1),
            nn.ConvBNReLU2d(
                out_channels, out_channels, kernel_size=(1, 7), padding=(0, 3)
            ),
            nn.ConvBNReLU2d(
                out_channels, out_channels, kernel_size=(7, 1), padding=(3, 0)
            ),
            nn.ConvBNReLU2d(
                out_channels, out_channels, kernel_size=(1, 7), padding=(0, 3)
            ),
            nn.ConvBNReLU2d(out_channels, 192, kernel_size=(7, 1), padding=(3, 0)),
        )

        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, out_channels, kernel_size=1),
            nn.ConvBNReLU2d(
                out_channels, out_channels, kernel_size=(1, 7), padding=(0, 3)
            ),
            nn.ConvBNReLU2d(out_channels, 192, kernel_size=(7, 1), padding=(3, 0)),
        )

        self.branch3 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
        )

        self.branch4 = nn.ConvBNReLU2d(in_channels, 192, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class _InceptionModule_V2C(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1_stem = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 448, kernel_size=1),
            nn.ConvBNReLU2d(448, 384, kernel_size=3, padding=1),
        )
        self.branch1_left = nn.ConvBNReLU2d(
            384, 384, kernel_size=(1, 3), padding=(0, 1)
        )
        self.branch1_right = nn.ConvBNReLU2d(
            384, 384, kernel_size=(3, 1), padding=(1, 0)
        )

        self.branch2_stem = nn.ConvBNReLU2d(in_channels, 384, kernel_size=1)
        self.branch2_left = nn.ConvBNReLU2d(
            384, 384, kernel_size=(1, 3), padding=(0, 1)
        )
        self.branch2_right = nn.ConvBNReLU2d(
            384, 384, kernel_size=(3, 1), padding=(1, 0)
        )

        self.branch3 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
        )

        self.branch4 = nn.ConvBNReLU2d(in_channels, 320, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        branch1_stem = self.branch1_stem(x)
        branch2_stem = self.branch2_stem(x)

        branch1 = lucid.concatenate(
            [self.branch1_left(branch1_stem), self.branch1_right(branch1_stem)],
            axis=1,
        )
        branch2 = lucid.concatenate(
            [self.branch2_left(branch2_stem), self.branch2_right(branch2_stem)],
            axis=1,
        )
        branch3 = self.branch3(x)
        branch4 = self.branch4(x)

        return lucid.concatenate([branch1, branch2, branch3, branch4], axis=1)


class _InceptionReduce_V2A(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 96, kernel_size=3, padding=1),
            nn.ConvBNReLU2d(96, 96, kernel_size=3, stride=2),
        )

        self.branch2 = nn.ConvBNReLU2d(in_channels, 384, kernel_size=3, stride=2)
        self.branch3 = nn.MaxPool2d(kernel_size=3, stride=2)

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x)], axis=1
        )


class _InceptionReduce_V2B(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
            nn.ConvBNReLU2d(192, 192, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(192, 192, kernel_size=(7, 1), padding=(3, 0)),
            nn.ConvBNReLU2d(192, 192, kernel_size=3, stride=2),
        )

        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
            nn.ConvBNReLU2d(192, 320, kernel_size=3, stride=2),
        )

        self.branch3 = nn.MaxPool2d(kernel_size=3, stride=2)

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x)], axis=1
        )


class Inception_V3(Inception):
    def __init__(
        self,
        num_classes: int = 1000,
        use_aux: bool = True,
        dropout_prob: float = 0.5,
    ) -> None:
        super().__init__(num_classes, use_aux)
        in_channels = 3

        self.conv1 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 32, kernel_size=3, stride=2),
            nn.ConvBNReLU2d(32, 32, kernel_size=3),
            nn.ConvBNReLU2d(32, 64, kernel_size=3, padding=1),
        )
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2)

        self.conv2 = nn.Sequential(
            nn.ConvBNReLU2d(64, 80, kernel_size=3),
            nn.ConvBNReLU2d(80, 192, kernel_size=3, stride=2),
            nn.ConvBNReLU2d(192, 288, kernel_size=3, padding=1),
        )

        self.incep_3 = nn.Sequential(
            _InceptionModule_V2A(288),
            _InceptionModule_V2A(288),
            _InceptionModule_V2A(288),
        )
        self.incep_red1 = _InceptionReduce_V2A(288)

        self.incep_4 = nn.Sequential(
            _InceptionModule_V2B(768, 128),
            _InceptionModule_V2B(768, 160),
            _InceptionModule_V2B(768, 160),
            _InceptionModule_V2B(768, 160),
            _InceptionModule_V2B(768, 192),
        )
        self.incep_red2 = _InceptionReduce_V2B(768)

        if use_aux:
            self.aux = _InceptionAux(768, num_classes, pool_size=(5, 5))
        else:
            self.aux = None

        self.incep_5 = nn.Sequential(
            _InceptionModule_V2C(1280), _InceptionModule_V2C(2048)
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=dropout_prob)
        self.fc = nn.Linear(2048, num_classes)

    def forward(self, x: Tensor) -> Tuple[Tensor, Optional[Tensor]]:
        x = self.maxpool(self.conv1(x))
        x = self.conv2(x)

        x = self.incep_3(x)
        x = self.incep_red1(x)
        x = self.incep_4(x)

        if self.aux is not None and self.training:
            aux = self.aux(x)
        else:
            aux = None

        x = self.incep_red2(x)
        x = self.avgpool(self.incep_5(x))

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x, aux if self.aux is not None else x


class _InceptionStem_V4(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.conv = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 32, kernel_size=3, stride=2),
            nn.ConvBNReLU2d(32, 32, kernel_size=3),
            nn.ConvBNReLU2d(32, 64, kernel_size=3, padding=1),
        )

        self.branch1_pool = nn.MaxPool2d(kernel_size=3, stride=2)
        self.branch1_conv = nn.ConvBNReLU2d(64, 96, kernel_size=3, stride=2)

        self.branch2_left = nn.Sequential(
            nn.ConvBNReLU2d(160, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 96, kernel_size=3),
        )
        self.branch2_right = nn.Sequential(
            nn.ConvBNReLU2d(160, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 64, kernel_size=(7, 1), padding=(3, 0)),
            nn.ConvBNReLU2d(64, 64, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(64, 96, kernel_size=3),
        )

        self.branch3_conv = nn.ConvBNReLU2d(192, 192, kernel_size=3, stride=2)
        self.branch3_pool = nn.MaxPool2d(kernel_size=3, stride=2)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)

        x = lucid.concatenate([self.branch1_pool(x), self.branch1_conv(x)], axis=1)
        x = lucid.concatenate([self.branch2_left(x), self.branch2_left(x)], axis=1)
        x = lucid.concatenate([self.branch3_pool(x), self.branch3_conv(x)], axis=1)

        return x


class _InceptionModule_V4A(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, padding=1),
            nn.ConvBNReLU2d(in_channels, 96, kernel_size=1),
        )
        self.branch2 = nn.ConvBNReLU2d(in_channels, 96, kernel_size=1)

        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 96, kernel_size=3, padding=1),
        )

        self.branch4 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 96, kernel_size=3, padding=1),
            nn.ConvBNReLU2d(96, 96, kernel_size=3, padding=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class _InceptionModule_V4B(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, padding=1),
            nn.ConvBNReLU2d(in_channels, 128, kernel_size=1),
        )

        self.branch2 = nn.ConvBNReLU2d(in_channels, 384, kernel_size=1)
        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
            nn.ConvBNReLU2d(192, 224, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(224, 256, kernel_size=(7, 1), padding=(3, 0)),
        )

        self.branch4 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 64, kernel_size=1),
            nn.ConvBNReLU2d(64, 192, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(192, 224, kernel_size=(7, 1), padding=(3, 0)),
            nn.ConvBNReLU2d(224, 224, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(224, 256, kernel_size=(7, 1), padding=(3, 0)),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class _InceptionModule_V4C(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.Sequential(
            nn.AvgPool2d(kernel_size=3, padding=1),
            nn.ConvBNReLU2d(in_channels, 256, kernel_size=1),
        )
        self.branch2 = nn.ConvBNReLU2d(in_channels, 256, kernel_size=1)

        self.branch3 = nn.ConvBNReLU2d(in_channels, 384, kernel_size=1)
        self.branch3_left = nn.ConvBNReLU2d(
            384, 256, kernel_size=(1, 3), padding=(0, 1)
        )
        self.branch3_right = nn.ConvBNReLU2d(
            384, 256, kernel_size=(3, 1), padding=(1, 0)
        )

        self.branch4 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 384, kernel_size=1),
            nn.ConvBNReLU2d(384, 448, kernel_size=(1, 3), padding=(0, 1)),
            nn.ConvBNReLU2d(448, 512, kernel_size=(3, 1), padding=(1, 0)),
        )
        self.branch4_left = nn.ConvBNReLU2d(
            512, 256, kernel_size=(3, 1), padding=(1, 0)
        )
        self.branch4_right = nn.ConvBNReLU2d(
            512, 256, kernel_size=(1, 3), padding=(0, 1)
        )

    def forward(self, x: Tensor) -> Tensor:
        x1 = self.branch1(x)
        x2 = self.branch2(x)

        x3 = self.branch3(x)
        x3l = self.branch3_left(x3)
        x3r = self.branch3_right(x3)

        x4 = self.branch4(x)
        x4l = self.branch4_left(x4)
        x4r = self.branch4_right(x4)

        return lucid.concatenate([x1, x2, x3l, x3r, x4l, x4r], axis=1)


class _InceptionReduce_V4A(nn.Module):
    def __init__(self, in_channels: int, k: int, l: int, m: int, n: int) -> None:
        super().__init__()

        self.branch1 = nn.MaxPool2d(kernel_size=3, stride=2)
        self.branch2 = nn.ConvBNReLU2d(in_channels, n, kernel_size=3, stride=2)

        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, k, kernel_size=1),
            nn.ConvBNReLU2d(k, l, kernel_size=3, padding=1),
            nn.ConvBNReLU2d(l, m, kernel_size=3, stride=2),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x)],
            axis=1,
        )


class _InceptionReduce_V4B(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.branch1 = nn.MaxPool2d(kernel_size=3, stride=2)
        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
            nn.ConvBNReLU2d(192, 192, kernel_size=3, stride=2),
        )

        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 256, kernel_size=1),
            nn.ConvBNReLU2d(256, 256, kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(256, 320, kernel_size=(7, 1), padding=(3, 0)),
            nn.ConvBNReLU2d(320, 320, kernel_size=3, stride=2),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x)],
            axis=1,
        )


class Inception_V4(Inception):
    def __init__(self, num_classes: int = 1000, use_aux: bool = True):
        super().__init__(num_classes, use_aux)
        in_channels = 3

        modules = []
        modules.append(_InceptionStem_V4(in_channels))

        for _ in range(4):
            modules.append(_InceptionModule_V4A(384))
        modules.append(_InceptionReduce_V4A(384, k=192, l=224, m=256, n=384))

        for _ in range(7):
            modules.append(_InceptionModule_V4B(1024))
        modules.append(_InceptionReduce_V4B(1024))

        for _ in range(3):
            modules.append(_InceptionModule_V4C(1536))

        self.conv = nn.Sequential(*modules)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=0.8)
        self.fc = nn.Linear(1536, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


@register_model
def inception_v1(
    num_classes: int = 1000,
    use_aux: bool = True,
    **kwargs,
) -> Inception:
    return Inception_V1(num_classes, use_aux, **kwargs)


@register_model
def inception_v3(
    num_classes: int = 1000,
    use_aux: bool = True,
    dropout_prob: float = 0.5,
    **kwargs,
) -> Inception:
    return Inception_V3(num_classes, use_aux, dropout_prob, **kwargs)


@register_model
def inception_v4(num_classes: int = 1000, **kwargs) -> Inception:
    return Inception_V4(num_classes, None, **kwargs)



from functools import partial
from typing import Literal, Type
from dataclasses import dataclass, field

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["CvT", "cvt_13", "cvt_21", "cvt_w24"]


def _to_2tuple(val: int | float) -> tuple[int | float, ...]:
    return (val, val)


class _QuickGELU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        return x * F.sigmoid(1.702 * x)


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: Type[nn.Module] = nn.GELU,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class _ConvAttention(nn.Module):
    def __init__(
        self,
        dim_in: int,
        dim_out: int,
        num_heads: int,
        qkv_bias: bool = False,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        method: Literal["dw_bn", "avg", "lin"] = "dw_bn",
        kernel_size: int = 3,
        stride_kv: int = 1,
        stride_q: int = 1,
        padding_kv: int = 1,
        padding_q: int = 1,
        with_cls_token: bool = True,
        **kwargs,
    ) -> None:
        super().__init__()
        self.stride_kv = stride_kv
        self.stride_q = stride_q
        self.dim = dim_out
        self.num_heads = num_heads

        self.scale = dim_out**-0.5
        self.with_cls_token = with_cls_token

        self.conv_proj_q = self._build_projection(
            dim_in,
            kernel_size,
            padding_q,
            stride_q,
            method="lin" if method == "avg" else method,
        )
        self.conv_proj_k = self._build_projection(
            dim_in, kernel_size, padding_kv, stride_kv, method
        )
        self.conv_proj_v = self._build_projection(
            dim_in, kernel_size, padding_kv, stride_kv, method
        )

        self.proj_q = nn.Linear(dim_in, dim_out, bias=qkv_bias)
        self.proj_k = nn.Linear(dim_in, dim_out, bias=qkv_bias)
        self.proj_v = nn.Linear(dim_in, dim_out, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim_out, dim_out)
        self.proj_drop = nn.Dropout(proj_drop)

    def _build_projection(
        self,
        dim_in: int,
        kernel_size: int,
        padding: int,
        stride: int,
        method: str,
    ) -> nn.Sequential | None:
        match method:
            case "dw_bn":
                proj = nn.Sequential(
                    nn.Conv2d(
                        dim_in,
                        dim_in,
                        kernel_size,
                        stride,
                        padding,
                        groups=dim_in,
                        bias=False,
                    ),
                    nn.BatchNorm2d(dim_in),
                    nn.Rearrange("b c h w -> b (h w) c"),
                )
            case "avg":
                proj = nn.Sequential(
                    nn.AvgPool2d(kernel_size, stride, padding),
                    nn.Rearrange("b c h w -> b (h w) c"),
                )
            case "lin":
                proj = None
            case _:
                raise ValueError(f"unknown method ({method}).")

        return proj

    def forward_conv(self, x: Tensor, h: int, w: int) -> tuple[Tensor, Tensor, Tensor]:
        if self.with_cls_token:
            cls_token, x = lucid.split(x, [1, h * w], axis=1)

        x = lucid.einops.rearrange(x, "b (h w) c -> b c h w", h=h, w=w)

        qkv = []
        for conv_proj in [self.conv_proj_q, self.conv_proj_k, self.conv_proj_v]:
            if conv_proj is not None:
                _proj = conv_proj(x)
            else:
                _proj = lucid.einops.rearrange(x, "b c h w -> b (h w) c")
            qkv.append(_proj)

        q, k, v = tuple(qkv)

        if self.with_cls_token:
            q = lucid.concatenate([cls_token, q], axis=1)
            k = lucid.concatenate([cls_token, k], axis=1)
            v = lucid.concatenate([cls_token, v], axis=1)

        return q, k, v

    def forward(self, x: Tensor, h: int, w: int) -> Tensor:
        if not (
            self.conv_proj_q is None
            and self.conv_proj_k is None
            and self.conv_proj_k is None
        ):
            q, k, v = self.forward_conv(x, h, w)

        q = self.proj_q(q)
        k = self.proj_k(k)
        v = self.proj_v(v)

        q = lucid.einops.rearrange(
            q, "b t (h d) -> b h t d", h=self.num_heads, d=q.shape[-1] // self.num_heads
        )
        k = lucid.einops.rearrange(
            k, "b t (h d) -> b h t d", h=self.num_heads, d=k.shape[-1] // self.num_heads
        )
        v = lucid.einops.rearrange(
            v, "b t (h d) -> b h t d", h=self.num_heads, d=v.shape[-1] // self.num_heads
        )

        attn_score = (q @ k.mT) * self.scale
        attn = F.softmax(attn_score, axis=-1)
        attn = self.attn_drop(attn)

        x = attn @ v
        x = lucid.einops.rearrange(x, "b h t d -> b t (h d)")

        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class _ConvTransformerBlock(nn.Module):
    def __init__(
        self,
        dim_in: int,
        dim_out: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        **kwargs,
    ) -> None:
        super().__init__()
        self.with_cls_token = kwargs["with_cls_token"]

        self.norm1 = norm_layer(dim_in)
        self.attn = _ConvAttention(
            dim_in, dim_out, num_heads, qkv_bias, attn_drop, drop, **kwargs
        )

        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim_out)

        mlp_hidden_dim = int(dim_out * mlp_ratio)
        self.mlp = _MLP(dim_out, mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x: Tensor, h: int, w: int) -> Tensor:
        res = x
        x = self.norm1(x)
        attn = self.attn(x, h, w)

        x = res + self.drop_path(attn)
        x += self.drop_path(self.mlp(self.norm2(x)))

        return x


class _ConvEmbed(nn.Module):
    def __init__(
        self,
        patch_size: int = 7,
        in_channels: int = 3,
        embed_dim: int = 64,
        stride: int = 4,
        padding: int = 2,
        norm_layer: Type[nn.Module] | None = None,
    ) -> None:
        super().__init__()
        patch_size = _to_2tuple(patch_size)
        self.patch_size = patch_size

        self.proj = nn.Conv2d(
            in_channels,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=padding,
        )
        self.norm = norm_layer(embed_dim) if norm_layer is not None else None

    def forward(self, x: Tensor) -> Tensor:
        x = self.proj(x)
        _, _, H, W = x.shape
        x = lucid.einops.rearrange(x, "b c h w -> b (h w) c")

        if self.norm is not None:
            x = self.norm(x)

        x = lucid.einops.rearrange(x, "b (h w) c -> b c h w", h=H, w=W)
        return x


class _VisionTransformer(nn.Module):
    def __init__(
        self,
        patch_size: int = 16,
        patch_stride: int = 16,
        patch_padding: int = 0,
        in_channels: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        **kwargs,
    ) -> None:
        super().__init__()
        self.num_features = self.embed_dim = embed_dim
        self.rearrange = None

        self.patch_embed = _ConvEmbed(
            patch_size,
            in_channels,
            stride=patch_stride,
            padding=patch_padding,
            embed_dim=embed_dim,
            norm_layer=norm_layer,
        )

        with_cls_token = kwargs["with_cls_token"]
        if with_cls_token:
            self.cls_token = nn.Parameter(lucid.zeros(1, 1, embed_dim))
        else:
            self.cls_token = None

        self.pos_drop = nn.Dropout(drop_rate)
        dpr = [x.item() for x in lucid.linspace(0, drop_path_rate, depth)]

        blocks = []
        for j in range(depth):
            blocks.append(
                _ConvTransformerBlock(
                    dim_in=embed_dim,
                    dim_out=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[j],
                    act_layer=act_layer,
                    norm_layer=norm_layer,
                    **kwargs,
                )
            )
        self.blocks = nn.ModuleList(blocks)

        if self.cls_token is not None:
            nn.init.normal(self.cls_token, std=0.02)

    def forward(self, x: Tensor) -> Tensor:
        x = self.patch_embed(x)
        B, _, H, W = x.shape

        x = lucid.einops.rearrange(x, "b c h w -> b (h w) c")

        cls_tokens = None
        if self.cls_token is not None:
            cls_tokens = self.cls_token.repeat(B, axis=0)
            x = lucid.concatenate([cls_tokens, x], axis=1)

        x = self.pos_drop(x)
        for block in self.blocks:
            x = block(x, H, W)

        if self.cls_token is not None:
            cls_tokens, x = lucid.split(x, [1, H * W], axis=1)

        x = lucid.einops.rearrange(x, "b (h w) c -> b c h w", h=H, w=W)

        return x, cls_tokens


@dataclass
class CvTSpec:
    num_stages: int
    patch_size: list[int]
    patch_stride: list[int]
    patch_padding: list[int]
    dim_embed: list[int]
    num_heads: list[int]
    depth: list[int]

    mlp_ratio: list[float] = field(default_factory=lambda: [4.0, 4.0, 4.0])
    attn_drop_rate: list[float] = field(default_factory=lambda: [0.0, 0.0, 0.0])
    drop_rate: list[float] = field(default_factory=lambda: [0.0, 0.0, 0.0])
    drop_path_rate: list[float] = field(default_factory=lambda: [0.0, 0.0, 0.1])

    qkv_bias: list[bool] = field(default_factory=lambda: [True, True, True])
    cls_token: list[bool] = field(default_factory=lambda: [False, False, True])
    pos_embed: list[bool] = field(default_factory=lambda: [False, False, False])

    qkv_proj_method: list[str] = field(
        default_factory=lambda: ["dw_bn", "dw_bn", "dw_bn"]
    )

    kernel_qkv: list[int] = field(default_factory=lambda: [3, 3, 3])
    padding_kv: list[int] = field(default_factory=lambda: [1, 1, 1])
    stride_kv: list[int] = field(default_factory=lambda: [2, 2, 2])
    padding_q: list[int] = field(default_factory=lambda: [1, 1, 1])
    stride_q: list[int] = field(default_factory=lambda: [1, 1, 1])


class CvT(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,
        num_classes: int = 1000,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        spec: CvTSpec | None = None,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.num_stages = spec.num_stages

        for i in range(self.num_stages):
            kwargs = dict(
                patch_size=spec.patch_size[i],
                patch_stride=spec.patch_stride[i],
                patch_padding=spec.patch_padding[i],
                embed_dim=spec.dim_embed[i],
                depth=spec.depth[i],
                num_heads=spec.num_heads[i],
                mlp_ratio=spec.mlp_ratio[i],
                qkv_bias=spec.qkv_bias[i],
                drop_rate=spec.drop_rate[i],
                attn_drop_rate=spec.attn_drop_rate[i],
                drop_path_rate=spec.drop_path_rate[i],
                with_cls_token=spec.cls_token[i],
                method=spec.qkv_proj_method[i],
                kernel_size=spec.kernel_qkv[i],
                padding_q=spec.padding_q[i],
                padding_kv=spec.padding_kv[i],
                stride_kv=spec.stride_kv[i],
                stride_q=spec.stride_q[i],
            )

            stage = _VisionTransformer(
                in_channels=in_channels,
                act_layer=act_layer,
                norm_layer=norm_layer,
                **kwargs,
            )
            setattr(self, f"stage{i}", stage)

            in_channels = spec.dim_embed[i]

        dim_embed = spec.dim_embed[-1]
        self.norm = norm_layer(dim_embed)
        self.cls_token = spec.cls_token[-1]

        self.head = (
            nn.Linear(dim_embed, num_classes) if num_classes > 0 else nn.Identity()
        )
        nn.init.normal(self.head.weight, std=0.02)

    def forward(self, x: Tensor) -> Tensor:
        for i in range(self.num_stages):
            x, cls_tokens = getattr(self, f"stage{i}")(x)

        if self.cls_token:
            x = self.norm(cls_tokens)
            x = lucid.squeeze(x, axis=1)
        else:
            x = lucid.einops.rearrange(x, "b c h w -> b (h w) c")
            x = self.norm(x)
            x = lucid.mean(x, axis=1)

        x = self.head(x)
        return x


@register_model
def cvt_13(num_classes: int = 1000, **kwargs) -> CvT:
    spec = CvTSpec(
        num_stages=3,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        dim_embed=[64, 192, 384],
        num_heads=[1, 3, 6],
        depth=[1, 2, 10],
        **kwargs,
    )
    return CvT(
        in_channels=3,
        num_classes=num_classes,
        act_layer=_QuickGELU,
        norm_layer=partial(nn.LayerNorm, eps=1e-5),
        spec=spec,
    )


@register_model
def cvt_21(num_classes: int = 1000, **kwargs) -> CvT:
    spec = CvTSpec(
        num_stages=3,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        dim_embed=[64, 192, 384],
        num_heads=[1, 3, 6],
        depth=[1, 4, 16],
        **kwargs,
    )
    return CvT(
        in_channels=3,
        num_classes=num_classes,
        act_layer=_QuickGELU,
        norm_layer=partial(nn.LayerNorm, eps=1e-5),
        spec=spec,
    )


@register_model
def cvt_w24(num_classes: int = 1000, **kwargs) -> CvT:
    spec = CvTSpec(
        num_stages=3,
        patch_size=[7, 3, 3],
        patch_stride=[4, 2, 2],
        patch_padding=[2, 1, 1],
        dim_embed=[192, 768, 1024],
        num_heads=[3, 12, 16],
        depth=[2, 2, 20],
        drop_path_rate=[0.0, 0.0, 0.3],
        **kwargs,
    )
    return CvT(
        in_channels=3,
        num_classes=num_classes,
        act_layer=_QuickGELU,
        norm_layer=partial(nn.LayerNorm, eps=1e-5),
        spec=spec,
    )



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = [
    "ReLU",
    "ReLU6",
    "LeakyReLU",
    "ELU",
    "SELU",
    "GELU",
    "Sigmoid",
    "HardSigmoid",
    "Tanh",
    "Softmax",
    "Swish",
    "HardSwish",
]


class ReLU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.relu(input_)


class ReLU6(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.relu(lucid.clip(input_, 0, 6))


class LeakyReLU(nn.Module):
    def __init__(self, negative_slope: float = 0.01) -> None:
        super().__init__()
        self.negative_slope = negative_slope

    def forward(self, input_: Tensor) -> Tensor:
        return F.leaky_relu(input_, self.negative_slope)


class ELU(nn.Module):
    def __init__(self, alpha: float = 1.0) -> None:
        super().__init__()
        self.alpha = alpha

    def forward(self, input_: Tensor) -> Tensor:
        return F.elu(input_, self.alpha)


class SELU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.selu(input_)


class GELU(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.gelu(input_)


class Sigmoid(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.sigmoid(input_)


class HardSigmoid(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return ((input_ / 6.0) + 0.5).clip(0, 1)


class Tanh(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return F.tanh(input_)


class Softmax(nn.Module):
    def __init__(self, axis: int = -1) -> None:
        super().__init__()
        self.axis = axis

    def forward(self, input_: Tensor) -> Tensor:
        return F.softmax(input_, self.axis)


class Swish(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        return input_ * F.sigmoid(input_)


class HardSwish(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, input_: Tensor) -> Tensor:
        hard_sigmoid = ((input_ / 6.0) + 0.5).clip(0, 1)
        return input_ * hard_sigmoid



from typing import Iterable
import numpy as np

import lucid
import lucid.nn as nn
import lucid.optim as optim

from lucid.types import _OptimClosure, _Scalar
from lucid._tensor import Tensor
from lucid._backend.metal import mx


__all__ = ["RMSprop", "Rprop"]


class RMSprop(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1e-2,
        alpha: float = 0.99,
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        momentum: float = 0.0,
        centered: bool = False,
    ) -> None:
        defaults = dict(
            lr=lr,
            alpha=alpha,
            eps=eps,
            weight_decay=weight_decay,
            momentum=momentum,
            centered=centered,
        )
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            alpha = group.get("alpha", self.defaults["alpha"])
            eps = group.get("eps", self.defaults["eps"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])
            momentum = group.get("momentum", self.defaults["momentum"])
            centered = group.get("centered", self.defaults["centered"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0.0:
                    grad += weight_decay * param.data

                state = self.state[param]

                if len(state) == 0:
                    state["step"] = 0
                    state["square_avg"] = lucid.zeros_like(param).data
                    if momentum != 0.0:
                        state["momentum_buffer"] = lucid.zeros_like(param).data
                    if centered:
                        state["grad_avg"] = lucid.zeros_like(param).data

                state["step"] += 1

                square_avg = state["square_avg"]
                square_avg[:] = alpha * square_avg + (1 - alpha) * (grad**2)

                if centered:
                    grad_avg = state["grad_avg"]
                    grad_avg[:] = alpha * grad_avg + (1 - alpha) * grad
                    avg = square_avg - grad_avg**2
                else:
                    avg = square_avg

                denom = lucid.sqrt(avg + eps).data
                if momentum != 0.0:
                    buf = state["momentum_buffer"]
                    buf[:] = momentum * buf + grad / denom
                    update = buf
                else:
                    update = grad / denom

                param.data -= lr * update

        return loss


class Rprop(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1e-2,
        etas: tuple[_Scalar, _Scalar] = (0.5, 1.2),
        step_sizes: tuple[_Scalar, _Scalar] = (1e-6, 50.0),
    ) -> None:
        defaults = dict(lr=lr, etas=etas, step_sizes=step_sizes)
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            etaminus, etaplus = group.get("etas", self.defaults["etas"])
            step_min, step_max = group.get("step_sizes", self.defaults["step_sizes"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                state = self.state[param]

                if len(state) == 0:
                    state["step"] = 0
                    state["prev_grad"] = lucid.zeros_like(param).data
                    state["step_size"] = lucid.ones_like(param).data * lr

                state["step"] += 1

                step_size = state["step_size"]
                prev_grad = state["prev_grad"]

                sign_change = grad * prev_grad

                lib_ = np if param.is_cpu() else mx
                step_size = lib_.where(sign_change > 0, step_size * etaplus, step_size)
                step_size = lib_.where(sign_change < 0, step_size * etaminus, step_size)

                step_size[:] = lucid.clip(step_size, step_min, step_max).data

                grad = lib_.where(sign_change < 0, 0, grad)
                state["prev_grad"] = Tensor.copy_grad(grad)

                param.data -= lucid.sign(grad).data * step_size

        return loss



from typing import Literal
from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _Scalar

from lucid.nn.functional import (
    _activation,
    _attention,
    _linear,
    _conv,
    _pool,
    _drop,
    _norm,
    _loss,
    _util,
)


def linear(input_: Tensor, weight: Tensor, bias: Tensor | None = None) -> Tensor:
    return _linear.linear(input_, weight, bias)


def bilinear(
    input_1: Tensor, input_2: Tensor, weight: Tensor, bias: Tensor | None = None
) -> Tensor:
    return _linear.bilinear(input_1, input_2, weight, bias)


def relu(input_: Tensor) -> Tensor:
    return _activation.relu(input_)


def leaky_relu(input_: Tensor, negative_slope: float = 0.01) -> Tensor:
    return _activation.leaky_relu(input_, negative_slope)


def elu(input_: Tensor, alpha: float = 1.0) -> Tensor:
    return _activation.elu(input_, alpha)


def selu(input_: Tensor) -> Tensor:
    return _activation.selu(input_)


def gelu(input_: Tensor) -> Tensor:
    return _activation.gelu(input_)


def sigmoid(input_: Tensor) -> Tensor:
    return _activation.sigmoid(input_)


def tanh(input_: Tensor) -> Tensor:
    return _activation.tanh(input_)


def softmax(input_: Tensor, axis: int = -1) -> Tensor:
    return _activation.softmax(input_, axis)


def unfold(
    input_: Tensor,
    filter_size: tuple[int, ...],
    stride: tuple[int, ...],
    padding: tuple[int, ...],
    dilation: tuple[int, ...],
) -> Tensor:
    return _conv.unfold(input_, filter_size, stride, padding, dilation)


def conv1d(
    input_: Tensor,
    weight: Tensor,
    bias: Tensor | None = None,
    stride: int | tuple[int, ...] = 1,
    padding: int | tuple[int, ...] = 0,
    dilation: int | tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    return _conv.conv1d(input_, weight, bias, stride, padding, dilation, groups)


def conv2d(
    input_: Tensor,
    weight: Tensor,
    bias: Tensor | None = None,
    stride: int | tuple[int, ...] = 1,
    padding: int | tuple[int, ...] = 0,
    dilation: int | tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    return _conv.conv2d(input_, weight, bias, stride, padding, dilation, groups)


def conv3d(
    input_: Tensor,
    weight: Tensor,
    bias: Tensor | None = None,
    stride: int | tuple[int, ...] = 1,
    padding: int | tuple[int, ...] = 0,
    dilation: int | tuple[int, ...] = 1,
    groups: int = 1,
) -> Tensor:
    return _conv.conv3d(input_, weight, bias, stride, padding, dilation, groups)


def avg_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return _pool.avg_pool1d(input_, kernel_size, stride, padding)


def avg_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return _pool.avg_pool2d(input_, kernel_size, stride, padding)


def avg_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return _pool.avg_pool3d(input_, kernel_size, stride, padding)


def max_pool1d(
    input_: Tensor,
    kernel_size: int | tuple[int] = 1,
    stride: int | tuple[int] = 1,
    padding: int | tuple[int] = 0,
) -> Tensor:
    return _pool.max_pool1d(input_, kernel_size, stride, padding)


def max_pool2d(
    input_: Tensor,
    kernel_size: int | tuple[int, int] = 1,
    stride: int | tuple[int, int] = 1,
    padding: int | tuple[int, int] = 0,
) -> Tensor:
    return _pool.max_pool2d(input_, kernel_size, stride, padding)


def max_pool3d(
    input_: Tensor,
    kernel_size: int | tuple[int, int, int] = 1,
    stride: int | tuple[int, int, int] = 1,
    padding: int | tuple[int, int, int] = 0,
) -> Tensor:
    return _pool.max_pool3d(input_, kernel_size, stride, padding)


def adaptive_avg_pool1d(input_: Tensor, output_size: int) -> Tensor:
    return _pool.adaptive_avg_pool1d(input_, output_size)


def adaptive_avg_pool2d(input_: Tensor, output_size: int | tuple[int, int]) -> Tensor:
    return _pool.adaptive_avg_pool2d(input_, output_size)


def adaptive_avg_pool3d(
    input_: Tensor, output_size: int | tuple[int, int, int]
) -> Tensor:
    return _pool.adaptive_avg_pool3d(input_, output_size)


def dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return _drop.dropout(input_, p, training)


def dropout1d(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return _drop.dropoutnd(input_, p, training)


def dropout2d(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return _drop.dropoutnd(input_, p, training)


def dropout3d(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return _drop.dropoutnd(input_, p, training)


def alpha_dropout(input_: Tensor, p: float = 0.5, training: bool = True) -> Tensor:
    return _drop.alpha_dropout(input_, p, training)


def drop_block(
    input_: Tensor, block_size: int, p: float = 0.1, eps: float = 1e-7
) -> Tensor:
    if input_.ndim != 4:
        raise ValueError("Only supports 4D tensors; Shape of '(N, C, H, W)'.")
    return _drop.drop_block(input_, block_size, p, eps)


def drop_path(input_: Tensor, p: float = 0.1, scale_by_keep: bool = True) -> Tensor:
    return _drop.drop_path(input_, p, scale_by_keep)


def normalize(
    input_: Tensor, ord: int = 2, axis: int = 1, eps: float = 1e-12
) -> Tensor:
    return _norm.normalize(input_, ord, axis, eps)


def batch_norm(
    input_: Tensor,
    running_mean: Tensor | None = None,
    running_var: Tensor | None = None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    return _norm.batch_norm(
        input_, running_mean, running_var, weight, bias, training, momentum, eps
    )


def layer_norm(
    input_: Tensor,
    normalized_shape: _ShapeLike,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    eps: float = 1e-5,
) -> Tensor:
    return _norm.layer_norm(input_, normalized_shape, weight, bias, eps)


def instance_norm(
    input_: Tensor,
    running_mean: Tensor | None = None,
    running_var: Tensor | None = None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    return _norm.instance_norm(
        input_, running_mean, running_var, weight, bias, training, momentum, eps
    )


def global_response_norm(
    input_: Tensor, gamma: Tensor, beta: Tensor, eps: float = 1e-6
) -> Tensor:
    return _norm.global_response_norm(input_, gamma, beta, eps)


_ReductionType = Literal["mean", "sum"]


def mse_loss(
    input_: Tensor, target: Tensor, reduction: _ReductionType | None = "mean"
) -> Tensor:
    return _loss.mse_loss(input_, target, reduction)


def binary_cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
) -> Tensor:
    return _loss.binary_cross_entropy(input_, target, weight, reduction, eps)


def cross_entropy(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
    eps: float = 1e-7,
) -> Tensor:
    return _loss.cross_entropy(input_, target, weight, reduction, eps)


def nll_loss(
    input_: Tensor,
    target: Tensor,
    weight: Tensor | None = None,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    return _loss.nll_loss(input_, target, weight, reduction)


def huber_loss(
    input_: Tensor,
    target: Tensor,
    delta: float = 1.0,
    reduction: _ReductionType | None = "mean",
) -> Tensor:
    return _loss.huber_loss(input_, target, delta, reduction)


_InterpolateType = Literal["bilinear", "nearest", "area"]


def interpolate(
    input_: Tensor,
    size: tuple[int, int],
    mode: _InterpolateType = "bilinear",
    align_corners: bool = False,
) -> Tensor:
    match mode:
        case "bilinear":
            return _util._interpolate_bilinear(input_, size, align_corners)
        case "nearest":
            return _util._interpolate_nearest(input_, size, align_corners)
        case "area":
            return _util._interpolate_area(input_, size, align_corners)
        case _:
            raise ValueError("Invalid interpolation type.")


def rotate(
    input_: Tensor, angle: float, center: tuple[_Scalar, _Scalar] | None = None
) -> Tensor:
    return _util.rotate(input_, angle, center)


def embedding(
    input_: Tensor,
    weight: Tensor,
    padding_idx: int | None = None,
    max_norm: float | None = None,
    norm_type: float = 2.0,
) -> Tensor:
    return _util.embedding(input_, weight, padding_idx, max_norm, norm_type)


def scaled_dot_product_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    attn_mask: Tensor | None = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    scale: _Scalar | None = None,
) -> Tensor:
    return _attention.scaled_dot_product_attention(
        query, key, value, attn_mask, dropout_p, is_causal, scale
    )



from typing import ClassVar, Literal, Type, override

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "SwinTransformer",
    "SwinTransformer_V2",
    "swin_tiny",
    "swin_small",
    "swin_base",
    "swin_large",
    "swin_v2_tiny",
    "swin_v2_small",
    "swin_v2_base",
    "swin_v2_large",
    "swin_v2_huge",
    "swin_v2_giant",
]


def _to_2tuple(val: int | float) -> tuple[int | float, ...]:
    return (val, val)


def window_partition(x: Tensor, window_size: int) -> Tensor:
    B, H, W, C = x.shape
    x = x.reshape(B, H // window_size, window_size, W // window_size, window_size, C)

    windows = x.swapaxes(2, 3).reshape(-1, window_size, window_size, C)
    return windows


def window_reverse(windows: Tensor, window_size: int, H: int, W: int) -> Tensor:
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.reshape(
        B, H // window_size, W // window_size, window_size, window_size, -1
    )
    x = x.swapaxes(2, 3).reshape(B, H, W, -1)
    return x


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: Type[nn.Module] = nn.GELU,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class _WindowAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        window_size: tuple[int, int],
        num_heads: int,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.window_size = window_size
        self.num_heads = num_heads

        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.relative_position_bias_table = nn.Parameter(
            lucid.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads)
        )
        nn.init.normal(self.relative_position_bias_table, std=0.02)

        coords_h = lucid.arange(self.window_size[0])
        coords_w = lucid.arange(self.window_size[1])

        coords = lucid.stack(lucid.meshgrid(coords_h, coords_w))
        coords_flatten = coords.reshape(coords.shape[0], -1)

        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
        relative_coords = relative_coords.transpose((1, 2, 0))

        relative_coords[:, :, 0] += self.window_size[0] - 1
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1

        relative_pos_index = relative_coords.sum(axis=-1)
        self.register_buffer("relative_pos_index", relative_pos_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)

        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.softmax = nn.Softmax(axis=-1)

    def forward(self, x: Tensor, mask: Tensor | None = None) -> Tensor:
        B_, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B_, N, 3, self.num_heads, C // self.num_heads)
            .transpose((2, 0, 3, 1, 4))
        )
        q, k, v = qkv[0], qkv[1], qkv[2]

        q *= self.scale
        attn = q @ k.mT

        relative_pos_bias = self.relative_position_bias_table[
            self.relative_pos_index.flatten().astype(lucid.Int)
        ].reshape(
            self.window_size[0] * self.window_size[1],
            self.window_size[0] * self.window_size[1],
            -1,
        )
        relative_pos_bias = relative_pos_bias.transpose((2, 0, 1))
        attn += relative_pos_bias.unsqueeze(axis=0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.reshape(B_ // nW, nW, self.num_heads, N, N)
            attn += mask.unsqueeze(axis=1).unsqueeze(axis=0)
            attn = attn.reshape(-1, self.num_heads, N, N)

        attn = self.softmax(attn)
        attn = self.attn_drop(attn)

        x = (attn @ v).swapaxes(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class _SwinTransformerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        input_res: tuple[int],
        num_heads: int,
        window_size: int = 7,
        shift_size: int = 0,
        mlp_ratio: int = 4,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.input_res = input_res
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio

        if min(self.input_res) <= self.window_size:
            self.shift_size = 0
            self.window_size = min(self.input_res)

        if not (0 <= self.shift_size < self.window_size):
            raise ValueError("shift_size must be in [0, window_size).")

        self.norm1 = norm_layer(dim)
        self.attn = _WindowAttention(
            dim,
            window_size=_to_2tuple(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )

        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)

        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = _MLP(dim, mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if self.shift_size > 0:
            H, W = self.input_res
            img_mask = lucid.zeros(1, H, W, 1)
            h_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )
            w_slices = (
                slice(0, -self.window_size),
                slice(-self.window_size, -self.shift_size),
                slice(-self.shift_size, None),
            )

            cnt = 0
            for h in h_slices:
                for w in w_slices:
                    img_mask[:, h, w, :] = cnt
                    cnt += 1

            mask_windows = window_partition(img_mask, self.window_size)
            mask_windows = mask_windows.reshape(-1, self.window_size * self.window_size)

            attn_mask = mask_windows.unsqueeze(axis=1) - mask_windows.unsqueeze(axis=2)
            attn_mask = attn_mask.masked_fill(attn_mask != 0, -100.0).masked_fill(
                attn_mask == 0, 0.0
            )
        else:
            attn_mask = None

        self.attn_mask: nn.Buffer
        self.register_buffer("attn_mask", attn_mask)

    def forward(self, x: Tensor) -> Tensor:
        H, W = self.input_res
        B, L, C = x.shape
        assert L == H * W, "wrong input feature size."

        shortcut = x
        x = self.norm1(x)
        x = x.reshape(B, H, W, C)

        if self.shift_size > 0:
            shifted_x = lucid.roll(
                x, shifts=(-self.shift_size, -self.shift_size), axis=(1, 2)
            )
            x_windows = window_partition(shifted_x, self.window_size)
        else:
            shifted_x = x
            x_windows = window_partition(shifted_x, self.window_size)

        x_windows = x_windows.reshape(-1, self.window_size * self.window_size, C)

        attn_windows = self.attn(x_windows, mask=self.attn_mask)
        attn_windows = attn_windows.reshape(-1, self.window_size, self.window_size, C)

        if self.shift_size > 0:
            shifted_x = window_reverse(attn_windows, self.window_size, H, W)
            x = lucid.roll(
                shifted_x, shifts=(self.shift_size, self.shift_size), axis=(1, 2)
            )
        else:
            shifted_x = window_reverse(attn_windows, self.window_size, H, W)
            x = shifted_x

        x = x.reshape(B, H * W, C)
        x = shortcut + self.drop_path(x)
        x += self.drop_path(self.mlp(self.norm2(x)))

        return x


class _PatchMerging(nn.Module):
    def __init__(
        self,
        input_res: tuple[int, int],
        dim: int,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        self.input_res = input_res
        self.dim = dim

        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x: Tensor) -> Tensor:
        H, W = self.input_res
        B, L, C = x.shape
        assert L == H * W, "wrong input feature size."
        assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

        x = x.reshape(B, H, W, C)
        x0 = x[:, 0::2, 0::2, :]
        x1 = x[:, 1::2, 0::2, :]
        x2 = x[:, 0::2, 1::2, :]
        x3 = x[:, 1::2, 1::2, :]

        x = lucid.concatenate([x0, x1, x2, x3], axis=-1)
        x = x.reshape(B, -1, 4 * C)

        x = self.norm(x)
        x = self.reduction(x)

        return x


class _BasicLayer(nn.Module):
    def __init__(
        self,
        dim: int,
        input_res: tuple[int, int],
        depth: int,
        num_heads: int,
        window_size: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float | list[float] = 0.0,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        downsample: nn.Module | None = None,
        _version: Literal["v1", "v2"] = "v1",
    ) -> None:
        super().__init__()
        self.dim = dim
        self.input_res = input_res
        self.depth = depth

        swin_block_dict = {"v1": _SwinTransformerBlock, "v2": _SwinTransformerBlock_V2}
        blocks = [
            swin_block_dict[_version](
                dim,
                input_res,
                num_heads,
                window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer,
            )
            for i in range(depth)
        ]
        self.blocks = nn.ModuleList(blocks)

        self.downsample: nn.Module | None
        if downsample is not None:
            self.downsample = downsample(input_res, dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x: Tensor) -> Tensor:
        for block in self.blocks:
            x = block(x)

        if self.downsample is not None:
            x = self.downsample(x)

        return x

    def _init_res_post_norm(self) -> None:
        for block in self.blocks:
            nn.init.constant(block.norm1.bias, 0)
            nn.init.constant(block.norm1.weight, 0)

            nn.init.constant(block.norm2.bias, 0)
            nn.init.constant(block.norm2.weight, 0)


class _PatchEmbed(nn.Module):
    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 4,
        in_channels: int = 3,
        embed_dim: int = 96,
        norm_layer: Type[nn.Module] | None = None,
    ) -> None:
        super().__init__()
        img_size = _to_2tuple(img_size)
        patch_size = _to_2tuple(patch_size)
        patches_res = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]

        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_res = patches_res
        self.num_patches = patches_res[0] * patches_res[1]

        self.in_channels = in_channels
        self.embed_dim = embed_dim

        self.proj = nn.Conv2d(
            in_channels, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x: Tensor) -> Tensor:
        _, _, H, W = x.shape
        assert H == self.img_size[0] and W == self.img_size[1], (
            f"Input image size ({H}*{W}) doesn't match model "
            + f"({self.img_size[0]}*{self.img_size[1]})."
        )
        x = self.proj(x)
        x = x.reshape(*x.shape[:2], -1).swapaxes(1, 2)

        if self.norm is not None:
            x = self.norm(x)

        return x


class SwinTransformer(nn.Module):
    _version: ClassVar[str] = "v1"

    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 4,
        in_channels: int = 3,
        num_classes: int = 1000,
        embed_dim: int = 96,
        depths: list[int] = [2, 2, 6, 2],
        num_heads: list[int] = [3, 6, 12, 24],
        windows_size: int = 7,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.1,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        abs_pos_emb: bool = False,
        patch_norm: bool = True,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.abs_pos_emb = abs_pos_emb
        self.patch_norm = patch_norm
        self.num_features = int(embed_dim * 2 ** (self.num_layers - 1))
        self.mlp_ratio = mlp_ratio

        self.patch_embed = _PatchEmbed(
            img_size,
            patch_size,
            in_channels,
            embed_dim,
            norm_layer=norm_layer if self.patch_norm else None,
        )
        num_patches = self.patch_embed.num_patches
        patches_res = self.patch_embed.patches_res
        self.patches_res = patches_res

        if self.abs_pos_emb:
            self.absolute_pos_emb = nn.Parameter(lucid.zeros(1, num_patches, embed_dim))
            nn.init.normal(self.absolute_pos_emb, std=0.02)

        self.pos_drop = nn.Dropout(drop_rate)
        dpr = [x.item() for x in lucid.linspace(0, drop_path_rate, sum(depths))]

        self.layers = nn.ModuleList()
        for i in range(self.num_layers):
            layer = _BasicLayer(
                dim=int(embed_dim * 2**i),
                input_res=(patches_res[0] // (2**i), patches_res[1] // (2**i)),
                depth=depths[i],
                num_heads=num_heads[i],
                window_size=windows_size,
                mlp_ratio=self.mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i]) : sum(depths[: i + 1])],
                norm_layer=norm_layer,
                downsample=_PatchMerging if i < self.num_layers - 1 else None,
                _version=self._version,
            )
            self.layers.append(layer)

        self.norm = norm_layer(self.num_features)
        self.avgpool = nn.AdaptiveAvgPool1d(1)
        self.head = (
            nn.Linear(self.num_features, num_classes)
            if num_classes > 0
            else nn.Identity()
        )

        self.apply(self._init_weights)

    def _init_weights(self, module: nn.Module) -> None:
        if isinstance(module, nn.Linear):
            nn.init.normal(module.weight, std=0.02)
            if module.bias is not None:
                nn.init.constant(module.bias, 0)

        elif isinstance(module, nn.LayerNorm):
            nn.init.constant(module.bias, 0)
            nn.init.constant(module.weight, 1.0)

    def forward(self, x: Tensor) -> Tensor:
        x = self.patch_embed(x)
        if self.abs_pos_emb:
            x += self.absolute_pos_emb
        x = self.pos_drop(x)

        for layer in self.layers:
            x = layer(x)

        x = self.norm(x)
        x = self.avgpool(x.swapaxes(1, 2))

        x = x.reshape(x.shape[0], -1)
        x = self.head(x)

        return x


class _WindowAttention_V2(nn.Module):
    def __init__(
        self,
        dim: int,
        window_size: tuple[int, int],
        num_heads: int,
        qkv_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.window_size = window_size
        self.num_heads = num_heads

        self.logit_scale = nn.Parameter(lucid.log(10 * lucid.ones(num_heads, 1, 1)))

        self.cpb_mlp = nn.Sequential(
            nn.Linear(2, 512), nn.ReLU(), nn.Linear(512, num_heads, bias=False)
        )

        rel_coords_h = lucid.arange(-(self.window_size[0] - 1), self.window_size[0])
        rel_coords_w = lucid.arange(-(self.window_size[1] - 1), self.window_size[1])

        rel_coords_table = (
            lucid.stack(lucid.meshgrid(rel_coords_h, rel_coords_w))
            .transpose((1, 2, 0))
            .unsqueeze(axis=0)
        )
        rel_coords_table[..., 0] /= self.window_size[0] - 1
        rel_coords_table[..., 1] /= self.window_size[1] - 1

        rel_coords_table *= 8
        rel_coords_table = (
            lucid.sign(rel_coords_table)
            * lucid.log2(lucid.abs(rel_coords_table) + 1.0)
            / lucid.log2(8)
        )

        self.rel_coords_table: nn.Buffer
        self.register_buffer("rel_coords_table", rel_coords_table)

        coords_h = lucid.arange(self.window_size[0])
        coords_w = lucid.arange(self.window_size[1])

        coords = lucid.stack(lucid.meshgrid(coords_h, coords_w))
        coords_flatten = coords.reshape(coords.shape[0], -1)

        rel_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
        rel_coords = rel_coords.transpose((1, 2, 0))

        rel_coords[:, :, 0] += self.window_size[0] - 1
        rel_coords[:, :, 1] += self.window_size[1] - 1
        rel_coords[:, :, 0] *= 2 * self.window_size[1] - 1

        rel_pos_index = rel_coords.sum(axis=-1)
        self.rel_pos_index: nn.Buffer
        self.register_buffer("rel_pos_index", rel_pos_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=False)
        if qkv_bias:
            self.q_bias = nn.Parameter(lucid.zeros(dim))
            self.v_bias = nn.Parameter(lucid.zeros(dim))
        else:
            self.q_bias = None
            self.v_bias = None

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.softmax = nn.Softmax(axis=-1)

    def forward(self, x: Tensor, mask: Tensor | None = None) -> Tensor:
        B_, N, C = x.shape
        qkv_bias = None

        if self.q_bias is not None:
            qkv_bias = lucid.concatenate(
                [self.q_bias, lucid.zeros_like(self.v_bias), self.v_bias]
            )
            qkv = F.linear(x, self.qkv.weight, qkv_bias)
            qkv = qkv.reshape(B_, N, 3, self.num_heads, -1).transpose((2, 0, 3, 1, 4))
            q, k, v = qkv[0], qkv[1], qkv[2]

            attn = F.normalize(q, axis=-1) @ F.normalize(k, axis=-1).mT
            logit_scale = lucid.clip(self.logit_scale, None, lucid.log(100.0).item())
            attn *= lucid.exp(logit_scale)

            rel_pos_bias_table = self.cpb_mlp(self.rel_coords_table).reshape(
                -1, self.num_heads
            )
            rel_pos_bias = rel_pos_bias_table[
                self.rel_pos_index.flatten().astype(lucid.Int)
            ].reshape(
                self.window_size[0] * self.window_size[1],
                self.window_size[0] * self.window_size[1],
                -1,
            )
            rel_pos_bias = rel_pos_bias.transpose((2, 0, 1))
            rel_pos_bias = 16 * F.sigmoid(rel_pos_bias)

            attn += rel_pos_bias.unsqueeze(axis=0)

            if mask is not None:
                nW = mask.shape[0]
                attn = attn.reshape(
                    B_ // nW, nW, self.num_heads, N, N
                ) + mask.unsqueeze(axis=1).unsqueeze(axis=0)

                attn = attn.reshape(-1, self.num_heads, N, N)
                attn = self.softmax(attn)
            else:
                attn = self.softmax(attn)

            attn = self.attn_drop(attn)

            x = (attn @ v).swapaxes(1, 2).reshape(B_, N, C)
            x = self.proj(x)
            x = self.proj_drop(x)

            return x


class _SwinTransformerBlock_V2(_SwinTransformerBlock):
    def __init__(
        self,
        dim: int,
        input_res: tuple[int, int],
        num_heads: int,
        window_size: int = 7,
        shift_size: int = 0,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer: Type[nn.Module] = nn.GELU,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__(
            dim,
            input_res,
            num_heads,
            window_size,
            shift_size,
            mlp_ratio,
            qkv_bias,
            qk_scale,
            drop,
            attn_drop,
            drop_path,
            act_layer,
            norm_layer,
        )
        self.attn = _WindowAttention_V2(
            dim,
            window_size=_to_2tuple(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )

    @override
    def forward(self, x: Tensor) -> Tensor:
        H, W = self.input_res
        B, L, C = x.shape
        assert L == H * W, "wrong input feature size."

        shortcut = x
        x = x.reshape(B, H, W, C)

        if self.shift_size > 0:
            shifted_x = x.roll((-self.shift_size, -self.shift_size), axis=(1, 2))
        else:
            shifted_x = x

        x_windows = window_partition(shifted_x, self.window_size)
        x_windows = x_windows.reshape(-1, self.window_size * self.window_size, C)

        attn_windows = self.attn(x_windows, mask=self.attn_mask)
        shifted_x = window_reverse(attn_windows, self.window_size, H, W)

        if self.shift_size > 0:
            x = shifted_x.roll((self.shift_size, self.shift_size), axis=(1, 2))
        else:
            x = shifted_x

        x = x.reshape(B, H * W, C)
        x = shortcut + self.drop_path(self.norm1(x))
        x += self.drop_path(self.norm2(self.mlp(x)))

        return x


class SwinTransformer_V2(SwinTransformer):
    _version: ClassVar[str] = "v2"

    def __init__(
        self,
        img_size: int = 224,
        patch_size: int = 4,
        in_channels: int = 3,
        num_classes: int = 1000,
        embed_dim: int = 96,
        depths: list[int] = [2, 2, 6, 2],
        num_heads: list[int] = [3, 6, 12, 24],
        windows_size: int = 7,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_scale: float | None = None,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.1,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        abs_pos_emb: bool = False,
        patch_norm: bool = True,
    ) -> None:
        super().__init__(
            img_size,
            patch_size,
            in_channels,
            num_classes,
            embed_dim,
            depths,
            num_heads,
            windows_size,
            mlp_ratio,
            qkv_bias,
            qk_scale,
            drop_rate,
            attn_drop_rate,
            drop_path_rate,
            norm_layer,
            abs_pos_emb,
            patch_norm,
        )
        for layer in self.layers:
            layer._init_res_post_norm()


@register_model
def swin_tiny(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer:
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    return SwinTransformer(
        img_size,
        num_classes=num_classes,
        embed_dim=96,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_small(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer:
    depths = [2, 2, 18, 2]
    num_heads = [3, 6, 12, 24]
    return SwinTransformer(
        img_size,
        num_classes=num_classes,
        embed_dim=96,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_base(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer:
    depths = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    return SwinTransformer(
        img_size,
        num_classes=num_classes,
        embed_dim=128,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_large(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer:
    depths = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    return SwinTransformer(
        img_size,
        num_classes=num_classes,
        embed_dim=192,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_tiny(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 6, 2]
    num_heads = [3, 6, 12, 24]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=96,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_small(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 18, 2]
    num_heads = [3, 6, 12, 24]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=96,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_base(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 18, 2]
    num_heads = [4, 8, 16, 32]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=128,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_large(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=192,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_huge(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 18, 2]
    num_heads = [6, 12, 24, 48]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=352,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )


@register_model
def swin_v2_giant(
    img_size: int = 224, num_classes: int = 1000, **kwargs
) -> SwinTransformer_V2:
    depths = [2, 2, 42, 4]
    num_heads = [6, 12, 24, 48]
    return SwinTransformer_V2(
        img_size,
        num_classes=num_classes,
        embed_dim=512,
        depths=depths,
        num_heads=num_heads,
        **kwargs,
    )



from typing import Iterable

import lucid
import lucid.nn as nn
import lucid.optim as optim

from lucid.types import _OptimClosure, _Scalar
from lucid._tensor import Tensor


__all__ = ["Adamax", "Adagrad", "Adadelta"]


class Adamax(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 2e-3,
        betas: tuple[_Scalar, _Scalar] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
    ) -> None:
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
        )
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            beta1, beta2 = group.get("betas", self.defaults["betas"])
            eps = group.get("eps", self.defaults["eps"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0.0:
                    grad += weight_decay * param.data

                state = self.state[param]
                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = lucid.zeros_like(param).data
                    state["exp_inf"] = lucid.zeros_like(param).data

                state["step"] += 1
                step = state["step"]
                exp_avg = state["exp_avg"]
                exp_inf = state["exp_inf"]

                exp_avg[:] = beta1 * exp_avg + (1 - beta1) * grad
                exp_inf[:] = lucid.maximum(beta2 * exp_inf, lucid.abs(grad)).data

                bias_correct1 = 1 - beta1**step

                step_size = lr / bias_correct1
                param.data -= step_size * (exp_avg / (exp_inf + eps))

        return loss


class Adagrad(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1e-2,
        eps: float = 1e-10,
        weight_decay: float = 0.0,
        initial_accumulator_value: float = 0.0,
    ) -> None:
        defaults = dict(
            lr=lr,
            eps=eps,
            weight_decay=weight_decay,
            initial_accumulator_value=initial_accumulator_value,
        )
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            eps = group.get("eps", self.defaults["eps"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])
            initial_accumulator_value = group.get(
                "initial_accumulator_value",
                self.defaults["initial_accumulator_value"],
            )

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0.0:
                    grad += weight_decay * param.data

                state = self.state[param]
                if len(state) == 0:
                    state["step"] = 0
                    state["sum_sq_grad"] = (
                        lucid.ones_like(param).data * initial_accumulator_value
                    )

                state["step"] += 1
                sum_sq_grad = state["sum_sq_grad"]
                sum_sq_grad[:] += grad**2

                step_size = lr / (lucid.sqrt(sum_sq_grad) + eps).data
                param.data -= step_size * grad

        return loss


class Adadelta(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1.0,
        rho: float = 0.9,
        eps: float = 1e-6,
        weight_decay: float = 0.0,
    ) -> None:
        defaults = dict(lr=lr, rho=rho, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            rho = group.get("rho", self.defaults["rho"])
            eps = group.get("eps", self.defaults["eps"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0.0:
                    grad += weight_decay * param.data

                state = self.state[param]
                if len(state) == 0:
                    state["step"] = 0
                    state["sq_avg"] = lucid.zeros_like(param).data
                    state["accumulated_update"] = lucid.zeros_like(param).data

                state["step"] += 1
                sq_avg = state["sq_avg"]
                accumulated_update = state["accumulated_update"]

                sq_avg[:] = rho * sq_avg + (1 - rho) * (grad**2)
                update = (
                    lucid.sqrt(accumulated_update + eps) / lucid.sqrt(sq_avg + eps)
                ).data * grad

                accumulated_update[:] = rho * accumulated_update + (1 - rho) * (
                    update**2
                )
                param.data -= lr * update

        return loss



import pandas as pd
import numpy as np
import openml
import math

from typing import SupportsIndex, Tuple, ClassVar

import lucid
from lucid._tensor import Tensor

from ._base import DatasetBase


__all__ = ["CIFAR10", "CIFAR100"]


class CIFAR10(DatasetBase):
    OPENML_ID: ClassVar[int] = 40927

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "CIFAR10.csv", index=False)

        except Exception as e:
            raise RuntimeError(f"Failed to download the CIFAR-10 dataset. Error: {e}")

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        csv_path = self.root / "CIFAR10.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"CIFAR-10 dataset CSV file not found at {csv_path}. "
                + "Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int32)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 3, 32, 32)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            images = lucid.to_tensor(images, dtype=lucid.Float32)
            labels = lucid.to_tensor(labels, dtype=lucid.Int32)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index].reshape(-1, 3, 32, 32)
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label


class CIFAR100(DatasetBase):
    OPENML_ID: ClassVar[int] = 41983

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "CIFAR100.csv", index=False)

        except Exception as e:
            raise RuntimeError(f"Failed to download the CIFAR-100 dataset. Error: {e}")

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        csv_path = self.root / "CIFAR100.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"CIFAR-100 dataset CSV file not found at {csv_path}. "
                + "Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int32)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 3, 32, 32)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            images = lucid.to_tensor(images, dtype=lucid.Float32)
            labels = lucid.to_tensor(labels, dtype=lucid.Int32)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index].reshape(-1, 3, 32, 32)
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label



from typing import Literal

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

from .inception import _InceptionStem_V4, _InceptionReduce_V4A


__all__ = ["InceptionResNet", "inception_resnet_v1", "inception_resnet_v2"]


class InceptionResNet(nn.Module):
    def __init__(self, num_classes: int) -> None:
        super().__init__()
        self.num_classes = num_classes

        self.stem: nn.Module
        self.conv: nn.Sequential
        self.fc: nn.Sequential


class _InceptionResStem(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.conv = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 32, kernel_size=3, stride=2),
            nn.ConvBNReLU2d(32, 32, kernel_size=3),
            nn.ConvBNReLU2d(32, 64, kernel_size=3, padding=1),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.ConvBNReLU2d(64, 80, kernel_size=1, padding=0),
            nn.ConvBNReLU2d(80, 192, kernel_size=3),
            nn.ConvBNReLU2d(192, 256, kernel_size=3, stride=2),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class _InceptionResModule_A(nn.Module):
    def __init__(self, in_channels: int, version: Literal["v1", "v2"]) -> None:
        super().__init__()

        if version == "v1":
            cfg = [32, 32, 256]
        elif version == "v2":
            cfg = [48, 64, 384]
        else:
            raise ValueError("Invalid version.")

        self.branch1 = nn.ConvBNReLU2d(in_channels, 32, kernel_size=1)
        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 32, kernel_size=1),
            nn.ConvBNReLU2d(32, 32, kernel_size=3, padding=1),
        )
        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 32, kernel_size=1),
            nn.ConvBNReLU2d(32, cfg[0], kernel_size=3, padding=1),
            nn.ConvBNReLU2d(cfg[0], cfg[1], kernel_size=3, padding=1),
        )

        self.conv_linear = nn.Conv2d(32 + 32 + cfg[1], cfg[2], kernel_size=1)
        self.relu = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        residual = x

        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        branch3 = self.branch3(x)
        out = lucid.concatenate([branch1, branch2, branch3], axis=1)

        out = self.conv_linear(out)
        out = out + residual
        out = self.relu(out)

        return out


class _InceptionResModule_B(nn.Module):
    def __init__(self, in_channels: int, version: Literal["v1", "v2"]) -> None:
        super().__init__()

        if version == "v1":
            cfg = [128, 128, 128, 896]
        elif version == "v2":
            cfg = [192, 160, 192, 1152]
        else:
            raise ValueError("Invalid version.")

        self.branch1 = nn.ConvBNReLU2d(in_channels, cfg[0], kernel_size=1)
        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 128, kernel_size=1),
            nn.ConvBNReLU2d(128, cfg[1], kernel_size=(1, 7), padding=(0, 3)),
            nn.ConvBNReLU2d(cfg[1], cfg[2], kernel_size=(7, 1), padding=(3, 0)),
        )

        self.conv_linear = nn.Conv2d(cfg[0] + cfg[2], cfg[3], kernel_size=1)
        self.relu = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        residual = x

        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        out = lucid.concatenate([branch1, branch2], axis=1)

        out = self.conv_linear(out)
        out = out + residual
        out = self.relu(out)

        return out


class _InceptionResModule_C(nn.Module):
    def __init__(self, in_channels: int, version: Literal["v1", "v2"]) -> None:
        super().__init__()

        if version == "v1":
            cfg = [192, 192, 1792]
        elif version == "v2":
            cfg = [224, 256, 2048]
        else:
            raise ValueError("Invalid version.")

        self.branch1 = nn.ConvBNReLU2d(in_channels, 192, kernel_size=1)
        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 192, kernel_size=1),
            nn.ConvBNReLU2d(192, cfg[0], kernel_size=(1, 3), padding=(0, 1)),
            nn.ConvBNReLU2d(cfg[0], cfg[1], kernel_size=(3, 1), padding=(1, 0)),
        )

        self.conv_linear = nn.Conv2d(192 + cfg[1], in_channels, kernel_size=1)
        self.relu = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        residual = x

        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        out = lucid.concatenate([branch1, branch2], axis=1)

        out = self.conv_linear(out)
        out = out + residual
        out = self.relu(out)

        return out


class _InceptionResReduce(nn.Module):
    def __init__(self, in_channels: int, version: Literal["v1", "v2"]) -> None:
        super().__init__()

        if version == "v1":
            cfg = [256, 256, 256]
        elif version == "v2":
            cfg = [288, 288, 320]
        else:
            raise ValueError("Invalid version.")

        self.branch1 = nn.MaxPool2d(kernel_size=3, stride=2)

        self.branch2 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 256, kernel_size=1),
            nn.ConvBNReLU2d(256, 384, kernel_size=3, stride=2),
        )

        self.branch3 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 256, kernel_size=1),
            nn.ConvBNReLU2d(256, cfg[0], kernel_size=3, stride=2),
        )

        self.branch4 = nn.Sequential(
            nn.ConvBNReLU2d(in_channels, 256, kernel_size=1),
            nn.ConvBNReLU2d(256, cfg[1], kernel_size=3, padding=1),
            nn.ConvBNReLU2d(cfg[1], cfg[2], kernel_size=3, stride=2),
        )

    def forward(self, x: Tensor) -> Tensor:
        return lucid.concatenate(
            [self.branch1(x), self.branch2(x), self.branch3(x), self.branch4(x)],
            axis=1,
        )


class InceptionResNet_V1(InceptionResNet):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__(num_classes)
        in_channels = 3

        self.stem = _InceptionResStem(in_channels)

        modules = []
        for _ in range(5):
            modules.append(_InceptionResModule_A(256, version="v1"))
        modules.append(_InceptionReduce_V4A(256, k=192, l=192, m=256, n=384))

        for _ in range(10):
            modules.append(_InceptionResModule_B(896, version="v1"))
        modules.append(_InceptionResReduce(896, version="v1"))

        for _ in range(5):
            modules.append(_InceptionResModule_C(1792, version="v1"))

        self.conv = nn.Sequential(*modules)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=0.8)
        self.fc = nn.Linear(1792, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


class InceptionResNet_V2(InceptionResNet):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__(num_classes)
        in_channels = 3

        self.stem = _InceptionStem_V4(in_channels)

        modules = []
        for _ in range(5):
            modules.append(_InceptionResModule_A(384, version="v2"))
        modules.append(_InceptionReduce_V4A(384, k=256, l=256, m=384, n=384))

        for _ in range(10):
            modules.append(_InceptionResModule_B(1152, version="v2"))
        modules.append(_InceptionResReduce(1152, version="v2"))

        for _ in range(5):
            modules.append(_InceptionResModule_C(2144, version="v2"))

        self.conv = nn.Sequential(*modules)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=0.8)
        self.fc = nn.Linear(2144, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


@register_model
def inception_resnet_v1(num_classes: int = 1000, **kwargs) -> InceptionResNet:
    return InceptionResNet_V1(num_classes, **kwargs)


@register_model
def inception_resnet_v2(num_classes: int = 1000, **kwargs) -> InceptionResNet:
    return InceptionResNet_V2(num_classes, **kwargs)



from typing import Callable
from functools import partial
import math

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "MaxViT",
    "maxvit_tiny",
    "maxvit_small",
    "maxvit_base",
    "maxvit_large",
    "maxvit_xlarge",
]


def _to_2tuple(val: int | float) -> tuple[int | float, ...]:
    return (val, val)


class _MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_feature: int | None = None,
        out_features: int | None = None,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] | None = None,
        bias: bool = True,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_feature = hidden_feature or in_features

        bias = _to_2tuple(bias)
        drop_probs = _to_2tuple(drop)

        self.fc1 = nn.Linear(in_features, hidden_feature, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.norm = (
            norm_layer(hidden_feature) if norm_layer is not None else nn.Identity()
        )

        self.fc2 = nn.Linear(hidden_feature, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x: Tensor) -> Tensor:
        x = self.drop1(self.act(self.fc1(x)))
        x = self.drop2(self.fc2(self.norm(x)))
        return x


class _SqueezeExcite(nn.Module):
    def __init__(
        self,
        in_channels: int,
        rd_ratio: float = 0.25,
        rd_channels: int | None = None,
        act_layer: type[nn.Module] = nn.ReLU,
        gate_layer: type[nn.Module] = nn.Sigmoid,
    ) -> None:
        super().__init__()
        if rd_channels is None:
            rd_channels = int(in_channels * rd_ratio)

        self.conv_reduce = nn.Conv2d(in_channels, rd_channels, kernel_size=1)
        self.act = act_layer()
        self.conv_expand = nn.Conv2d(rd_channels, in_channels, kernel_size=1)
        self.gate = gate_layer()

    def forward(self, x: Tensor) -> Tensor:
        x_se = x.mean(axis=(2, 3), keepdims=True)
        x_se = self.conv_reduce(x_se)
        x_se = self.act(x_se)
        x_se = self.conv_expand(x_se)

        return x * self.gate(x_se)


class _DepthwiseSeparableConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 1,
        dilation: int = 1,
        act_layer: type[nn.Module] = nn.ReLU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        drop_path_rate: float = 0.0,
    ) -> None:
        super().__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups=in_channels,
            bias=False,
        )
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        self.act = act_layer()
        self.bn = norm_layer(out_channels)
        self.drop_path = (
            nn.DropPath(drop_path_rate) if drop_path_rate > 0.0 else nn.Identity()
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.depthwise(x)
        x = self.pointwise(x)
        x = self.drop_path(self.act(self.bn(x)))

        return x


class _MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        downscale: bool = False,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        drop_path: float = 0.0,
    ) -> None:
        super().__init__()
        self.drop_path_rate = drop_path
        if not downscale:
            assert (
                in_channels == out_channels
            ), "in/out channels must be equal when downscale=True."
        self.in_channels = in_channels
        self.main_path = nn.Sequential(
            norm_layer(in_channels),
            nn.Conv2d(in_channels, in_channels, kernel_size=1),
            _DepthwiseSeparableConv(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=2 if downscale else 1,
                act_layer=act_layer,
                drop_path_rate=drop_path,
            ),
            _SqueezeExcite(out_channels, rd_ratio=0.25),
            nn.Conv2d(out_channels, out_channels, kernel_size=1),
        )
        self.skip_path = (
            nn.Sequential(
                nn.MaxPool2d(kernel_size=2, stride=2),
                nn.Conv2d(in_channels, out_channels, kernel_size=1),
            )
            if downscale
            else nn.Identity()
        )
        self.drop = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        out = self.main_path(x)
        out = self.drop(out)
        out += self.skip_path(x)

        return out


def _window_partition(x: Tensor, window_size: tuple[int, int] = (7, 7)) -> Tensor:
    B, C, H, W = x.shape
    windows = x.reshape(
        B, C, H // window_size[0], window_size[0], W // window_size[1], window_size[1]
    )
    windows = windows.transpose((0, 2, 4, 3, 5, 1)).reshape(-1, *window_size, C)
    return windows


def _window_reverse(
    windows: Tensor,
    original_size: tuple[int, int],
    window_size: tuple[int, int] = (7, 7),
) -> Tensor:
    H, W = original_size
    B = int(windows.shape[0] / (H * W / window_size[0] / window_size[1]))

    out = windows.reshape(B, H // window_size[0], W // window_size[1], *window_size, -1)
    out = out.transpose((0, 5, 1, 3, 2, 4)).reshape(B, -1, H, W)
    return out


def _grid_partition(x: Tensor, grid_size: tuple[int, int] = (7, 7)) -> Tensor:
    B, C, H, W = x.shape
    grid = x.reshape(
        B, C, grid_size[0], H // grid_size[0], grid_size[1], W // grid_size[1]
    )
    grid = grid.transpose((0, 3, 5, 2, 4, 1)).reshape(-1, *grid_size, C)
    return grid


def _grid_reverse(
    grid: Tensor, original_size: tuple[int, int], grid_size: tuple[int, int] = (7, 7)
) -> Tensor:
    (H, W), C = original_size, grid.shape[-1]
    B = int(grid.shape[0] / (H * W / grid_size[0] / grid_size[1]))
    out = grid.reshape(B, H // grid_size[0], W // grid_size[1], *grid_size, C)
    out = out.transpose((0, 5, 3, 1, 4, 2)).reshape(B, C, H, W)
    return out


def _get_relative_position_index(win_h: int, win_w: int) -> Tensor:
    coords = lucid.stack(lucid.meshgrid(lucid.arange(win_h), lucid.arange(win_w)))
    coords_flatten = lucid.flatten(coords, axis=1)

    relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
    relative_coords = relative_coords.transpose((1, 2, 0))

    relative_coords[:, :, 0] += win_h - 1
    relative_coords[:, :, 1] += win_w - 1
    relative_coords[:, :, 0] *= win_w - 1

    return relative_coords.sum(axis=-1).astype(lucid.Int)


class _RelativeSelfAttention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int = 32,
        grid_window_size: tuple[int, int] = (7, 7),
        attn_drop: float = 0.0,
        drop: float = 0.0,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.num_heads = num_heads
        self.grid_window_size = grid_window_size

        self.scale = num_heads**-0.5
        self.attn_area = grid_window_size[0] * grid_window_size[1]

        self.qkv_mapping = nn.Linear(in_channels, 3 * in_channels)
        self.attn_drop = nn.Dropout(attn_drop)

        self.proj = nn.Linear(in_channels, in_channels)
        self.proj_drop = nn.Dropout(drop)
        self.softmax = nn.Softmax(axis=-1)

        self.relative_pos_bias_table = nn.Parameter(
            lucid.zeros(
                (2 * grid_window_size[0] - 1) * (2 * grid_window_size[1] - 1), num_heads
            )
        )
        nn.init.normal(self.relative_pos_bias_table, std=0.02)

        self.relative_pos_index: nn.Buffer
        self.register_buffer(
            "relative_pos_index", _get_relative_position_index(*grid_window_size)
        )

    def _get_relative_positional_bias(self) -> Tensor:
        rel_pos_bias = self.relative_pos_bias_table[
            self.relative_pos_index.reshape(-1)
        ].reshape(self.attn_area, self.attn_area, -1)

        rel_pos_bias = rel_pos_bias.transpose((2, 0, 1))
        return rel_pos_bias.unsqueeze(axis=0)

    def forward(self, x: Tensor) -> Tensor:
        B_, N, _ = x.shape
        qkv = (
            self.qkv_mapping(x)
            .reshape(B_, N, 3, self.num_heads, -1)
            .transpose((2, 0, 3, 1, 4))
        )
        q, k, v = qkv.chunk(3)
        q *= self.scale

        attn = self.softmax(q @ k.mT + self._get_relative_positional_bias())
        out = (attn @ v).swapaxes(1, 2).reshape(B_, N, -1)
        out = self.proj(out)
        out = self.proj_drop(out)

        return out


class _MaxViTTransformerBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        partition_func: Callable,
        reverse_func: Callable,
        num_heads: int = 32,
        grid_window_size: tuple[int, int] = (7, 7),
        attn_drop: float = 0.0,
        drop: float = 0.0,
        drop_path: float = 0.0,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        self.partition_func: Callable[[Tensor, tuple], Tensor] = partition_func
        self.reverse_func: Callable[[Tensor, tuple, tuple], Tensor] = reverse_func
        self.grid_window_size = grid_window_size

        self.norm_1 = norm_layer(in_channels)
        self.attention = _RelativeSelfAttention(
            in_channels, num_heads, grid_window_size, attn_drop, drop
        )
        self.drop_path = nn.DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.norm_2 = norm_layer(in_channels)
        self.mlp = _MLP(
            in_channels, int(mlp_ratio * in_channels), act_layer=act_layer, drop=drop
        )

    def forward(self, x: Tensor) -> Tensor:
        _, C, H, W = x.shape
        input_part = self.partition_func(x, self.grid_window_size)
        input_part = input_part.reshape(-1, math.prod(self.grid_window_size), C)

        out = input_part + self.drop_path(self.attention(self.norm_1(input_part)))
        out += self.drop_path(self.mlp(self.norm_2(out)))
        out = self.reverse_func(out, (H, W), self.grid_window_size)

        return out


class _MaxViTBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        downscale: bool = False,
        num_heads: int = 32,
        grid_window_size: tuple[int, int] = (7, 7),
        attn_drop: float = 0.0,
        drop: float = 0.0,
        drop_path: float = 0.0,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        norm_layer_tf: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        self.mb_conv = _MBConv(
            in_channels, out_channels, downscale, act_layer, norm_layer, drop_path
        )
        base_tf = partial(
            _MaxViTTransformerBlock,
            in_channels=out_channels,
            num_heads=num_heads,
            grid_window_size=grid_window_size,
            attn_drop=attn_drop,
            drop=drop,
            drop_path=drop_path,
            mlp_ratio=mlp_ratio,
            act_layer=act_layer,
            norm_layer=norm_layer_tf,
        )

        self.block_tf = base_tf(
            partition_func=_window_partition, reverse_func=_window_reverse
        )
        self.grid_tf = base_tf(
            partition_func=_grid_partition, reverse_func=_grid_reverse
        )

    def forward(self, x: Tensor) -> Tensor:
        out = self.mb_conv(x)
        out = self.block_tf(out)
        out = self.grid_tf(out)

        return out


class _MaxViTStage(nn.Module):
    def __init__(
        self,
        depth: int,
        in_channels: int,
        out_channels: int,
        num_heads: int = 32,
        grid_window_size: tuple[int, int] = (7, 7),
        attn_drop: float = 0.0,
        drop: float = 0.0,
        drop_path: list[float] | float = 0.0,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        norm_layer_tf: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        blocks = [
            _MaxViTBlock(
                in_channels=in_channels if idx == 0 else out_channels,
                out_channels=out_channels,
                downscale=idx == 0,
                num_heads=num_heads,
                grid_window_size=grid_window_size,
                attn_drop=attn_drop,
                drop=drop,
                drop_path=drop_path if isinstance(drop_path, float) else drop_path[idx],
                mlp_ratio=mlp_ratio,
                act_layer=act_layer,
                norm_layer=norm_layer,
                norm_layer_tf=norm_layer_tf,
            )
            for idx in range(depth)
        ]
        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: Tensor) -> Tensor:
        return self.blocks(x)


class MaxViT(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,
        depths: tuple[int, ...] = (2, 2, 5, 2),
        channels: tuple[int, ...] = (64, 128, 256, 512),
        num_classes: int = 1000,
        embed_dim: int = 64,
        num_heads: int = 32,
        grid_window_size: tuple[int, int] = (7, 7),
        attn_drop: float = 0.0,
        drop: float = 0.0,
        drop_path: float = 0.0,
        mlp_ratio: float = 4.0,
        act_layer: type[nn.Module] = nn.GELU,
        norm_layer: type[nn.Module] = nn.BatchNorm2d,
        norm_layer_tf: type[nn.Module] = nn.LayerNorm,
    ) -> None:
        super().__init__()
        assert len(depths) == len(
            channels
        ), "Channel dims must be given for each stage."

        self.num_classes = num_classes
        self.stem = nn.Sequential(
            nn.Conv2d(in_channels, embed_dim, kernel_size=3, stride=2, padding=1),
            act_layer(),
            nn.Conv2d(embed_dim, embed_dim, kernel_size=3, stride=1, padding=1),
            act_layer(),
        )

        drop_path = lucid.linspace(0, drop_path, sum(depths))
        stages = []
        for idx, (depth, channel) in enumerate(zip(depths, channels)):
            stages.append(
                _MaxViTStage(
                    depth=depth,
                    in_channels=embed_dim if idx == 0 else channels[idx - 1],
                    out_channels=channel,
                    num_heads=num_heads,
                    grid_window_size=grid_window_size,
                    attn_drop=attn_drop,
                    drop=drop,
                    drop_path=drop_path[sum(depths[:idx]) : sum(depths[: idx + 1])],
                    mlp_ratio=mlp_ratio,
                    act_layer=act_layer,
                    norm_layer=norm_layer,
                    norm_layer_tf=norm_layer_tf,
                )
            )

        self.stages = nn.ModuleList(stages)
        self.head = nn.Linear(channels[-1], num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        for stage in self.stages:
            x = stage(x)

        x = x.mean(axis=(2, 3))
        out = self.head(x)
        return out


@register_model
def maxvit_tiny(in_channels: int = 3, num_classes: int = 1000, **kwargs) -> MaxViT:
    return MaxViT(
        in_channels=in_channels,
        num_classes=num_classes,
        depths=(2, 2, 5, 2),
        channels=(64, 128, 256, 512),
        embed_dim=64,
        **kwargs
    )


@register_model
def maxvit_small(in_channels: int = 3, num_classes: int = 1000, **kwargs) -> MaxViT:
    return MaxViT(
        in_channels=in_channels,
        num_classes=num_classes,
        depths=(2, 2, 5, 2),
        channels=(96, 192, 384, 768),
        embed_dim=64,
        **kwargs
    )


@register_model
def maxvit_base(in_channels: int = 3, num_classes: int = 1000, **kwargs) -> MaxViT:
    return MaxViT(
        in_channels=in_channels,
        num_classes=num_classes,
        depths=(2, 6, 14, 2),
        channels=(96, 192, 384, 768),
        embed_dim=64,
        **kwargs
    )


@register_model
def maxvit_large(in_channels: int = 3, num_classes: int = 1000, **kwargs) -> MaxViT:
    return MaxViT(
        in_channels=in_channels,
        num_classes=num_classes,
        depths=(2, 6, 14, 2),
        channels=(128, 256, 512, 1024),
        embed_dim=128,
        **kwargs
    )


@register_model
def maxvit_xlarge(in_channels: int = 3, num_classes: int = 1000, **kwargs) -> MaxViT:
    return MaxViT(
        in_channels=in_channels,
        num_classes=num_classes,
        depths=(2, 6, 14, 2),
        channels=(192, 384, 768, 1536),
        embed_dim=192,
        **kwargs
    )
